#%% ##################################################################### Imports #####################################################################

# Numpy 
import numpy as np
import numpy.linalg as npla
# np.set_printoptions(threshold=np.inf) # print entire arrays
np.set_printoptions(linewidth=np.inf) # don't auto-wrap arrays, important for weight coefficients

# Matplotlib
import matplotlib.pyplot as plt

from tqdm import tqdm # progressbar

from abc import ABC, abstractmethod # ABstract Class stuff

from sklearn.model_selection import ParameterGrid # for grid search construction

import pandas as pd # DataFrame export of gridsearch results
pd.set_option("display.max_rows", 10000) # numbers of rows to display for tables

from scipy.signal import lfilter

##################################################################### Auxiliary Functions #####################################################################
# *********************************************************************** Plotter tool ********************************************************************
def PltFormat(Fig, Axis, Title=None, XLab=None, XLim=[], YLab=None, YLim=[], Legend =[], Show=True):
   '''Sets the passed arguments
   Fig and Axis are obtained with "Fig, Ax = plt.subplots(rows, cols)" - no arguments for a single plot
   Note 1: the Legend Argument a list must be passed even if a single element is to be set
   Note 2: The {X,Y}Lim arguments expect 2 Element lists, pass [] to let matplotlib decide
   '''
   Axis.grid(which="both");Fig.tight_layout();
   if (Title):  Axis.set_title(Title)
   if (XLab):   Axis.set_xlabel(XLab) # X Axis label
   if (XLim):   Axis.set_xlim(XLim) # if non-empty list, sets X Axis Limits
   if (YLab):   Axis.set_ylabel(YLab) # Y Axis label
   if (YLim):   Axis.set_ylim(YLim) # if non-empty list, sets Y Axis Limits
   if (Legend): Axis.legend(Legend) # if list non empty apply legend
   if (Show):   plt.show() # prevents the overlap of xlabel and title

# *********************************************************************** grid search/ cross validation ********************************************************************
def gridsearch(Params, x, d, Plot=True, Future=0, TestData = None):
   '''Performs a grid search for the filter given by "FilterType" on the given hyper-parameters.
      Dictionary must contain a "FilterType" entry and the minimum parameters for the concerned filters to work
      
      Returns:
      1) Hyper-parameters dict of the best performing Filter for filter creation (contains no FilterType nor MSE value, see Table)
      2) Pandas DataFrame of FilterType, hyperparameters and MSE'''
       
   # defensive programming: must be a) a dict b) containing lists and c) have an entry "FilterTypes" containing known filters
   if (not isinstance(Params, dict)): raise ValueError("Params must be a dictionary") # must be a dictionary
   
   for element in list(Params): # all passed parameters in the dict must be lists
      if (not isinstance(Params[element], list)): raise ValueError("Params dictionary must only contain lists")

   for Type in Params["FilterType"]: # list must be passed for ParameterGrid to work, check if no unknown filters are contained
      if (Type not in ["LMS", "pRLS", "KLMS"]): raise ValueError("Unknown Filter '" + Type +"' demanded, accepted types are LMS, pRLS, KLMS")
   # x and d are checked by member function adapt

   Results = {key : [] for key in Params.keys()} # create a dict of lists for upcomming parameters to be appended
   Results['MSE'] = [] # MSE is a supplementary entry not contained in the Params dict

   if(Plot): Fig, Ax = plt.subplots() # create figure

   # grid search
   ParamGrid = ParameterGrid(Params) # Parameter Grid construction
   # print("\n\nThe Parameter Grid has", len(ParamGrid)," elements\n\n")
   for GridPoint in tqdm(ParamGrid): # iterate over all parameter combinations

      # Add current settings to Results dict (before taking out FilterType)
      for key, value in GridPoint.items(): Results[key].append(value)

      # Legend formatting to clean up the plot: if entry exists format, irrespective of the filter type
      Temp = GridPoint.copy() # deep copy of the dict to prevent alias creation
      if("MatrixInit" in Temp): Temp["MatrixInit"] = '{:.3e}'.format(Temp["MatrixInit"]) #pRLS only
      if("p" in Temp): Temp["p"] = str(np.round(Temp["p"],4)) #pRLS only
      if("Mu" in Temp): Temp["Mu"] = '{:.3e}'.format(Temp["Mu"]) # LMS and KLMS
      if("KernelSize" in Temp): Temp["KernelSize"] = '{:.3e}'.format(Temp["KernelSize"])
      GridPointName = str(Temp)[1:-1].replace("'","") # eliminate first and last chars being {} and all '' and FilterType

      # create right Filter
      FilterType = GridPoint.pop("FilterType") # extract filter type of dict
      # construct the right Filter with remaining parameters, errors will be thrown if incorrect, so no verificatin necessary
      if   (FilterType == "LMS"):  Filter =  LMS(**GridPoint)
      elif (FilterType == "pRLS"): Filter = pRLS(**GridPoint)
      elif (FilterType == "KLMS"): Filter = KLMS(**GridPoint)

      # fit filter
      MSError = Filter.adapt(x, d, Plot=False, Future=Future)[2] # take only MSE-array (exclude prediction error, filtered Output and weights)
      if (type(TestData) is np.ndarray):
         PrError = Filter.filter(np.pad(TestData, (Future,0)))[Future:] - TestData
         MSError = np.divide(np.cumsum(PrError**2), np.arange(1, len(PrError)+1) ) # MSE = cumulative error / 1-based iteration number

      if(Plot): Ax.plot(MSError, label=GridPointName) # plot MSE immediately to avoid having to store it
      Results['MSE'].append(MSError[-1]) # only store last MSE
      # print(GridPointName + " finished\n")

   if(Plot):
      Ax.legend(ncol=3, loc="best")# print the legend created during the loop
      PltFormat(Fig, Ax, "Mean Squared Error", "Time in Samples", [0, len(x)], "Amplitude", [0, np.max(MSError)], [], Show = False)

   # Find best and format Hyper-parameters
   ResultsDF = pd.DataFrame(Results) # create table for a clear result display or csv export
   BestHyperParams = ResultsDF.iloc[ResultsDF["MSE"].idxmin(),:].to_dict() # transform the entry with the minimum MSE into a dict
   print("\n\nResults Table:\n", ResultsDF, "\n\nBest Performing Filter type and parameters:\n", BestHyperParams, "\n\n")
   del BestHyperParams["FilterType"], BestHyperParams["MSE"] # allows BestHyperParams to be used to create a Filter with those hyperparameters

   return(BestHyperParams, ResultsDF)

# *********************************************************************** LMS converge ********************************************************************
def LMSconverge(Input, NTaps, Mu = None, Print = True):
   '''Helper function making prophecies about LMS filter convergence
   Input:
   - Input: Signal the LMS is to be adapted on
   - NTaps: Number of tap the filter has (order+1), if several NTaps are tested by a gridsearch, the largest should be passed
   - Mu: (optional) desired stepsize to be tested for convergeance
   - Print (optional) if the results are to be displayed on the console

   Output:
   - Converge: bool estimating if the filter converges (if no Mu is passed, False is automatically returned)
   - ConvTime: expected convergence time of the LMS filter (-1 if no Mu is passed, being an impossible result)
   - MaxMu:  Maximum stepsize Mu at which the filter is expected to converge
   - InputPower: Power of the given Input, used by the internal computations
   '''
   InputPower = np.mean(Input**2) # assumes real signal, else np.abs is needed
   MaxMu = 2/(NTaps*InputPower)
   if (Print): print("\nThe input-power is", np.round(InputPower,3),"allowing a maximal converging mu of", np.round(MaxMu,4), "\n")

   if(Mu == None): return(False, -1, MaxMu, InputPower) # use [2:] in receiving environment
   else: 
      Converge = Mu < MaxMu
      ConvTime = np.ceil(1/ (2*Mu *InputPower))
      return(Converge, ConvTime, MaxMu, InputPower)
# ############################################################################## AF Class #####################################################################
# *********************************************************************** Adaptive Filter class *****************************************************
class AdaptiveFilter(ABC): # abstract class, no C-Tor

# ------------------------------------------------------------------------- W() ---------------------------------------------------------------------
   def W(self):
      '''Returns the Weights for LMS and RLS and the Centerset and their Coefficients for KLMS'''
      # RLS and LMS require their weight vector to be flipped to be in the correct order (output only, no internal effect)
      if   (self.Type == "LMS"):  return(np.flip(self._LMS__W))
      elif (self.Type == "pRLS"): return(np.flip(self._pRLS__W))
      elif (self.Type == "KLMS"): return(self._KLMS__W, self.a)

# ----------------------------------------------------------------------- NTaps() ---------------------------------------------------------------------
   def NTaps(self):
      '''Returns the Number of taps for LMS and pRLS and the windowsize (samples taken for Kernel computation) for KLMS'''
      if(self.Type in ["LMS", "pRLS"]): return (len(self.W())) # (W() is required since private vars aren't accessible from the superclass)
      elif(self.Type == "KLMS"):        return (self.WindowSize) # needed for adapation (equivalent to order, so to say)

# ----------------------------------------------------------------------- print() ---------------------------------------------------------------------
   @abstractmethod # force subclasses to overwrite, as all other parameters are class specific
   def print(self): print("\nFilter type:", self.Type, "\nNumber of taps (=order+1) :", self.NTaps(), "\nWeights:", self.W())
      
# --------------------------------------------------------------------- iteration() -----------------------------------------------------------------
   @abstractmethod # force subclasses to overwrite
   def iteration(self): pass # nothing in common

# ------------------------------------------------------------------------ adapt() ---------------------------------------------------------------------
   def adapt(self, x, d, Plot=True, Future=0):
      '''Adapt the filter to the given Input and output the filtered signal and goodness of fit measured and weights

      Inputs:
      x: Filter Input, comparison signal
      d: desired signal, signal to be compared to
      Plot: plots the adaption process in three parts: 1) output vs desired signal 2) MSE evolution 3) Prediction error evolution
      Future: Allows fitting for prediction. Gives the number of samples the filter should lookahead, zero for normal fitting

      Outputs:
      - OutSignal (Filter output Signal): Filtered signal 
      - PrError (Prediction error): predicted value - desired value at each iteration
      - MSError (Mean squared Error): average output error over entire adaptation
      - self.__W: Filter weights in case the filter is part of a hyper-parameter search
      '''
      
      # defensive programming : control the inputs
      if (x.ndim != 1): raise ValueError("Input Signal 'x' must be Mono")
      if (d.ndim != 1): raise ValueError("Desired Signal 'd' must be Mono")
      if (len(x) != len(d)) : raise ValueError("Both signals must be of the same length")

      # Input Padding
      InSignal = np.pad(x, (self.NTaps()-1+Future, 0) )
      # number of iterations is len(x), add self.NTaps()-1 to let filter "ring out" = end of input signal.
      # Zero padding is required since the first value x[0] requires [NTaps]-1 previous values for the FIR scalarproduct to be calculated
      # this additional length is not considered in the iterations (so no change for d[i])

      # Requirted Arrays:
      SigLen = len(x) # unpadded length, equiavant to the length of d
      OutSignal = np.full(SigLen, np.nan, dtype = np.float32) # init to nan to recognize unwrittten entries (no nan expected from multiplication)
      PrError = np.full(SigLen, np.nan, dtype = np.float32) # as above, Prediction Error
      
      # Actual filter iterations
      Taps = self.NTaps() # optimization avoiding calling two functions and one if/else in every iteration
      for i in range(SigLen): OutSignal[i], PrError[i] = self.iteration(InSignal[i:i+Taps], d[i]) # one iteration per signal sample
      
      MSError = np.divide(np.cumsum(PrError**2), np.arange(1, len(PrError)+1) ) # MSE = cumulative error / 1-based iteration number
      
      if(self.Type =="KLMS"): # cleans up the 0 initialisation
         self._KLMS__W = self._KLMS__W[1:] # way of getting acces to the private variable
         self.a = self.a[1:]

      if(Plot): # display the adaption process if desired
         Iter = range(SigLen)

         Fig, Ax = plt.subplots(3, 1) # 3 rows 1 column
         Fig.suptitle(self.Type + " with " + str(self.NTaps()) + " tap(s) and " + str(np.round(np.mean(x**2), 3)) + " Input Power")

         Ax[0].plot(Iter, d, Iter, OutSignal) # compare Filter Output with Desired Signal (ground truth)
         PltFormat(Fig, Ax[0], "Adaption", "Time in Samples", [0, SigLen], "Amplitude", [], ["Desired Signal", "Output Signal"], Show = False)

         Ax[1].plot(Iter, MSError) # Mean squared error
         PltFormat(Fig, Ax[1], "Mean Squared Error", "Time in Samples", [0, SigLen], "Amplitude", [0, np.max(MSError)], [], Show = False)

         Ax[2].plot(Iter, np.abs(PrError)) # momentaneous prediction error
         PltFormat(Fig, Ax[2], "Absolute Prediction Error", "Time in Samples", [0, SigLen], "Amplitude", [0, np.max(PrError)], [], Show = False)

         if (self.Type in ["LMS", "pRLS"]): print("\nFilter weight: ", self.W(), "\n") # KLMS is just an un-informative enormous list

      return (OutSignal, PrError, MSError, self.W()) # Return filter output and prediction and MS error tuple

# ----------------------------------------------------------------------- filter() ---------------------------------------------------------------------
   def filter(self, Input): 
      '''Apply the calling filter, once adapted.
     Single Argument: Input to apply filter on
     Returns filtered signal'''
      if(self.Type in ["LMS", "pRLS"]):
         if(np.array_equal(self.W(), np.zeros(self.NTaps()))): raise AssertionError("All weights = 0, thus the filter is uninitialized, use adapt first")
         return lfilter(self.W(), [1.0], Input) # use premade filter function

      elif(self.Type == "KLMS"): # exactly what adapt does but without updating weights
         if(np.array_equal(self.W()[0], np.zeros(self.NTaps()))): raise AssertionError("All weights = 0, thus the filter is uninitialized, use adapt first")

         SigLen = len(Input)
         InSignal = np.pad(Input, (self.NTaps()-1, 0) )
         OutSignal = np.full(SigLen, np.nan, dtype = np.float32)
         # transform lists into ararys once for all since they are static
         W = np.array(self._KLMS__W)
         a = np.array(self.a)

         Taps = self.NTaps() # optimization avoiding calling two functions and one if/else in every iteration
         for i in range(SigLen): OutSignal[i] = self.f(InSignal[i:i+Taps], W, a) # one iteration per signal sample
         return(OutSignal) # norming the signal /np.max(OutSignal)
         
         

# %% ############################################################################## pRLS Class #####################################################################
class pRLS(AdaptiveFilter):
   # class has no setters since NTaps and weights are not expected to be changed

# ----------------------------------------------------------------------- CTor() ---------------------------------------------------------------------
   def __init__(self, NTaps=20, p=0.995, MatrixInit=1000): # C-Tor
      self.Type = "pRLS" # set Filter Type
      if(p<=0 or p>1): raise AssertionError("forget factor must be in ]0;1]") # doesn't fail if p is not a scalar
      if(NTaps<0): raise AssertionError("filter NTaps must be >0")
      self.__W = np.zeros(int(NTaps)) # weightvector, force NTaps to be an integer, fails if W is incorrectly a vector
      self.InvR = MatrixInit*np.eye(NTaps) # large scalar matrix as default initialisation (full rank matrix)
      self.p = float(p) # fails if p is incorrectly a vector

# ----------------------------------------------------------------------- print() ---------------------------------------------------------------------
   def print(self):
      '''Print the entire content of the filter class'''
      super().print() # print all common parameters
      print("Forget factor p:", self.p, "\nInverted Autocorrelation matrix R:\n", self.InvR, "\n") # class specific params

# ----------------------------------------------------------------------- iteration() -----------------------------------------------------------------
   def iteration(self, x, d):
      z = np.full(len(x), np.nan, dtype=np.float32) # init z vector
      y = np.dot(x, self.__W) # 1. apriori filter output
      e = d - y # 2. apriori error
      z = (self.InvR @ x) / (self.p + x @ self.InvR @ x) # 3. filtered normalized vector
      self.InvR = (self.InvR - np.outer(z, x) @ self.InvR)/self.p # 4. actualize autocorr matrix (can also be done after 4.)
      self.__W += e*z
      return (y, e) # filter Output and prediction error

# ############################################################################## LMS Class #####################################################################
class LMS(AdaptiveFilter):
   # class has no setters since NTaps and weights are not expected to be changed

# ----------------------------------------------------------------------- C-Tor() -----------------------------------------------------------------------
   def __init__(self, NTaps=20, Mu=0.001): # C-Tor
      self.Type = "LMS" # set Filter Type
      if(Mu<=0): raise AssertionError("Stepsize must be >0") # doesn't fail if Mu is not a scalar
      if(NTaps<0): raise AssertionError("Filter NTaps must be >0")
      self.__W = np.zeros(int(NTaps)) # weightvector, force NTaps to be an integer, fails if NTaps is incorrectly a vector
      self.Mu = float(Mu) # fails if Mu is incorrectly a vector

# ----------------------------------------------------------------------- print() ---------------------------------------------------------------------
   def print(self): # display all filter information
      '''Print the entire content of the filter class'''
      super().print() # print all common AF parameterss
      print("Stepsize Mu:", self.Mu, "\n")
   
# ----------------------------------------------------------------------- iteration () -----------------------------------------------------------------
   def iteration(self, x, d):
      y = np.dot(self.__W, x)  # 1. Filter output
      e = d-y # 2. apriori error
      self.__W += self.Mu*e*x # actualize filter coefficients
      return (y, e) # return filter output and error

# ############################################################################## NC-KLMS Class #####################################################################
# ----------------------------------------------------------------------- Kernel Function ---------------------------------------------------------------------
def GaussianKernel (x, Y, h=10):
   ''' vectorized Gaussian Kernels for entire array, optimized for KLMS
   No size verifications are done, to avoid slowing down function, numpy throws errors accordingly
   Inputs:
   x: 1D (W,)-sized numpy array of the current window to be processed
   Y: 2D Numpy array of N (W,)-sized numpy arrays (shape = (N,W))
   h: Gaussian Kernel width

   Output: 1D (N,)-sized scalar array, resulting from the Gaussian Kernel
   '''
   return np.exp( npla.norm(x-Y, axis=1)**2 / (-2*h**2) )


class KLMS(AdaptiveFilter):
   # Note that the Centers set is named W for coherence with the other filters adn to simplify inheritance

# ----------------------------------------------------------------------- C-Tor() ---------------------------------------------------------------------
   def __init__(self, Mu=0.1, WindowSize=1, Kernel = GaussianKernel, KernelSize=10.0, NC=False, NC_MinDist=0.01, NC_MinErr=0.01): # C-Tor
      self.Type = "KLMS" # set Filter Type

      if(Mu<=0): raise ValueError("Stepsize must be >0") # doesn't fail if Mu is not a scalar
      if(WindowSize<=0): raise AssertionError("Windowsize must be >0")

      self.Mu = float(Mu) # fails if Mu is incorrectly a vector
      self.WindowSize = WindowSize
      self.K = Kernel # kernel function pointer
      self.KernelSize = float(KernelSize) # passes argument for the Kernel function, error if vector
      # filled at every iteration, will thus __W and a have the same length, init to 0 for np.min(), makes no difference in sum, deleted afer adaption
      self.__W = [np.zeros(WindowSize)] # Center list (x[i] values)
      self.a = [0] # coefficient list (Mu* error)
      self.NC = bool(NC) # novelty criterion, force to be bool

      if(NC): # create entries only if needed
         self.NC_MinDist = NC_MinDist
         self.NC_MinError = NC_MinErr

# ----------------------------------------------------------------------- print() ---------------------------------------------------------------------
   def print(self):
      '''Print the entire content of the filter class'''
      super().print() # print all common AF parameters
      print("Stepsize Mu:", self.Mu,"\nWindow-size:", self.WindowSize, "\nGaussian Kernel size:", self.KernelSize)
      if(self.NC==True): print("Platt's Novelty criterion:","\nMinimal Distance:", self.NC_MinDist, "\nMinimal Error", self.NC_MinError) 

# ----------------------------------------------------------------------- f() -----------------------------------------------------------------
   def f(self, x, W, a): return np.dot(np.array(a), self.K(x, W, self.KernelSize) ) # KernelSum/filter output, vectorized
      
# ----------------------------------------------------------------------- iteration() -----------------------------------------------------------------
   def iteration(self, x, d):
      W = np.array(self.__W) # transform List into np.array for vectorization
      f_ = self.f(x, W, self.a)
      e = d - f_ # error

      # Platt's novelty criterion (heuristic to keep the dictionary small)
      # If no NC - add automatically, if novelty criterion + 2 NC-critera met - add, else skip addition
      if( (self.NC == False) or ( self.NC==True and (np.min(npla.norm(x-W, axis=1)) > self.NC_MinDist) and (np.abs(e) > self.NC_MinError) )): 
         self.__W.append(x)
         self.a.append(self.Mu * e)
      return(f_, e) # filter output and error